from aiohttp import ClientSession
from requests import Session
from chiefpay.constants import BASE_URL, Endpoints
from typing import Union


class BaseClient:
    """
    Base class for interacting with the payment system.
    """
    def __init__(self, api_key: str, base_url: str = BASE_URL):
        """
        Initialize the client.

        Parameters:
            api_key (str): API key for authentication.
            base_url (str): Base URL for the API endpoints.
        """
        self.api_key = api_key
        self.base_url = base_url
        self.headers = {
            "Accept": "application/json",
            "X-Api-Key": self.api_key
        }
        self.session: Union[Session, ClientSession] = None


    def _init_session(self):
        raise NotImplementedError

    def _get_url(self, endpoint: Endpoints):
        if not self.session:
            self.session = self._init_session()

        url = self.base_url + endpoint.value
        return url


import aiohttp
from typing import Dict, Optional
from chiefpay.base import BaseClient
from chiefpay.constants import Endpoints
from chiefpay.exceptions import APIError, HTTPError, ManyRequestsError
from chiefpay.types import Rate, Wallet, Invoice, InvoicesHistory, TransactionsHistory, Transaction
from chiefpay.utils import Utils

from asyncio import sleep


class AsyncClient(BaseClient):
    """
    Client for making asynchronous requests to the payment system API.
    """

    def _init_session(self):
        return aiohttp.ClientSession(headers=self.headers)

    async def _request(self, method: str, path: str, max_retries: int = 3, **kwargs):
        url = self._get_url(path)
        for attempt in range(max_retries):
            try:
                async with self.session.request(method, url, **kwargs) as response:
                    return await self._handle_response(response)
            except ManyRequestsError:
                if attempt == max_retries - 1:
                    raise HTTPError(429, f"Too Many Requests after {max_retries} attempts")
                continue

    async def _get_request(self, path: str, params: Optional[Dict] = {}, max_retries: int = 3):
        params = {k: v for k, v in params.items() if v is not None}

        return await self._request("GET", path, max_retries, params=params)

    async def _post_request(self, path: str, json: Optional[Dict] = {}, max_retries: int = 3):
        json = {k: v for k, v in json.items() if v is not None}

        return await self._request("POST", path, max_retries, json=json)

    @staticmethod
    async def _handle_response(response: aiohttp.ClientResponse):
        if response.status == 429:
            headers = response.headers
            retry = int(headers.get('Retry-After-ms', '3000')) / 1000
            await sleep(retry)
            raise ManyRequestsError()

        if not (200 <= response.status < 300):
            text = await response.text()
            raise HTTPError(response.status, text)
        try:
            data = await response.json()
            return data.get('data')
        except ValueError:
            raise APIError("Invalid JSON response")


    async def get_rates(self) -> list[Rate]:
        """
        Asynchronously retrieves the current exchange rates.

        Returns:
             Rate DTO: The exchange rate data.
        """
        rates = await self._get_request(Endpoints.rates)
        return [Rate(**rate) for rate in rates]

    async def get_invoice(self, id: Optional[str] = None, order_id: Optional[str] = None) -> Invoice:
        """
        Asynchronously retrieves information about a specific invoice.

        Parameters:
            id (str): The invoice ID.
            order_id (str): The order ID.

        Returns:
             Invoice DTO: The invoice data.
        """
        params = {
            "id": id,
            "orderId": order_id
        }

        response_data = await self._get_request(Endpoints.invoice, params)
        return Invoice(**response_data)


    async def get_invoices(self, from_date: str, to_date: Optional[str] = None, limit: int = 100) -> InvoicesHistory:
        """
        Asynchronously retrieves a list of invoices within a specified date range.
        Args:
            from_date (str): The start date for the invoice history in 'YYYY-MM-DD' format.
            to_date (Optional[str], optional): The end date for the invoice history in 'YYYY-MM-DD' format. Defaults to None.
            limit (int, optional): The maximum number of invoices to retrieve. Defaults to 100.
        Returns:
            InvoicesHistory: An object containing the list of invoices and the total count of invoices.
        Raises:
            ValueError: If the date format for `from_date` or `to_date` is invalid.
        """
        Utils.validate_date(from_date)
        if to_date:
            Utils.validate_date(to_date)

        params = {
            "fromDate": from_date, 
            "limit": limit, 
            "toDate": to_date
        }

        response_data = await self._get_request(Endpoints.invoices_history, params)
        invoices = [Invoice(**data) for data in response_data.get('invoices')]
        return InvoicesHistory(
            invoices=invoices,
            totalCount=response_data.get('totalCount')
        )

    async def get_transactions(self, from_date: str, to_date: Optional[str] = None, limit: int = 100) -> TransactionsHistory:
        """
        Asynchronously retrieves transaction history within a specified date range.
        Args:
            from_date (str): The start date for the transaction history in 'YYYY-MM-DD' format.
            to_date (Optional[str], optional): The end date for the transaction history in 'YYYY-MM-DD' format. Defaults to None.
            limit (int, optional): The maximum number of transactions to retrieve. Defaults to 100.
        Returns:
            TransactionsHistory: An object containing the list of transactions and the total count.
        Raises:
            ValueError: If the date format for `from_date` or `to_date` is invalid.
        """
        Utils.validate_date(from_date)
        if to_date:
            Utils.validate_date(to_date)

        params = {
            "fromDate": from_date, 
            "limit": limit, 
            "toDate": to_date
        }

        response_data = await self._get_request(Endpoints.transactions_history, params)
        transactions = [Transaction(**data) for data in response_data.get('transactions')]
        return TransactionsHistory(
            transactions=transactions,
            totalCount=response_data.get('totalCount')
        )

    async def get_wallet(self, id: Optional[str] = None, order_id: Optional[str] = None) -> Wallet:
        """
        Retrieve wallet information based on wallet ID or order ID.
        Args:
            id (Optional[str]): The ID of the wallet to retrieve.
            order_id (Optional[str]): The order ID associated with the wallet to retrieve.
        Returns:
            Wallet: An instance of the Wallet class containing the retrieved wallet information.
        Raises:
            ValueError: If neither `id` nor `order_id` is provided.
        """
        params = {
            "id": id,
            "orderId": order_id
        }

        response_data = await self._get_request(Endpoints.wallet, params)
        return Wallet(**response_data)


    async def create_invoice(
        self,
        order_id: str,
        description: Optional[str] = None,
        amount: Optional[float] = None,
        currency: Optional[str] = "USD",
        fee_included: Optional[bool] = False,
        accuracy: Optional[float] = None,
        discount: Optional[float] = None,
    ) -> Invoice:
        """
        Asynchronously creates a new invoice.

        Parameters:
            order_id (str): The order ID.
            description (str): The invoice description.
            amount (float): The amount.
            currency (str): The currency.
            fee_included (bool): Whether the fee is included in the amount.
            accuracy (float): The accuracy level.
            discount (float): The discount.

        Returns:
             Invoice DTO: The created invoice data.
        """
        data = {
            "orderId": order_id,
            "description": description,
            "amount": amount,
            "currency": currency,
            "feeIncluded": fee_included,
            "accuracy": accuracy,
            "discount": discount,
        }

        response_data = await self._post_request(Endpoints.invoice, json=data)
        return Invoice(**response_data)

    async def create_wallet(self, order_id: str) -> Wallet:
        """
        Asynchronously creates a new wallet.

        Parameters:
            order_id (str): The order ID.

        Returns:
             Wallet DTO: The created wallet data.
        """
        data = {
            "orderId": order_id
        }

        response_data = await self._post_request(Endpoints.wallet, json=data)
        return Wallet(**response_data)


    async def close(self):
        """
        Closes the asynchronous session.

        This should be called after all asynchronous requests are complete.
        """
        await self.session.close()


    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()



import requests
from typing import Dict, Optional
from time import sleep

from chiefpay.base import BaseClient
from chiefpay.constants import Endpoints
from chiefpay.exceptions import HTTPError, APIError, ManyRequestsError
from chiefpay.types import Rate, Wallet, Invoice, InvoicesHistory, TransactionsHistory, Transaction
from chiefpay.utils import Utils


class Client(BaseClient):
    """
    Client for making synchronous requests to the payment system API.
    """
    def _init_session(self):
        session = requests.session()
        session.headers.update(self.headers)
        return session

    def _request(self, method: str, path: str, max_retries: int = 3, **kwargs):
        url = self._get_url(path)
        for attempt in range(max_retries):
            response = self.session.request(method, url, **kwargs)
            try:
                return self._handle_response(response)
            except ManyRequestsError:
                if attempt == max_retries - 1:
                    raise HTTPError(429, f"Too Many Requests after {max_retries} attempts")
                continue

    def _get_request(self, path: str, params: Optional[Dict] = None, max_retries: int = 3):
        return self._request("GET", path, max_retries, params=params)

    def _post_request(self, path: str, json: Optional[Dict] = None, max_retries: int = 3):
        return self._request("POST", path, max_retries, json=json)

    @staticmethod
    def _handle_response(response: requests.Response):
        if response.status_code == 429:
            headers = response.headers
            retry = int(headers.get('Retry-After-ms', '3000')) / 1000
            sleep(retry)
            raise ManyRequestsError()

        if not (200 <= response.status_code < 300):
            raise HTTPError(response.status_code, response.text)

        try:
            data = response.json()
            return data.get('data')
        except ValueError:
            raise APIError("Invalid JSON response")


    def get_rates(self) -> list[Rate]:
        """
        Retrieves the current exchange rates.

        Returns:
             Rate DTO: The exchange rate data.
        """
        response_data = self._get_request(Endpoints.rates)
        return [Rate(**rate) for rate in response_data]

    def get_invoice(self, id: Optional[str] = None, order_id: Optional[str] = None) -> Invoice:
        """
        Retrieves information about a specific invoice.

        Parameters:
            id (str): The invoice ID.
            order_id (str): The order ID.

        Returns:
             Invoice DTO: The invoice data.
        """
        if id:
            params = {"id": id}
        elif order_id:
            params = {"orderId": order_id}
        response_data = self._get_request(Endpoints.invoice, params)
        return Invoice(**response_data)

    def get_invoices(self, from_date: str, to_date: Optional[str] = None, limit: int = 100) -> InvoicesHistory:
        """
        Retrieves invoices history within a given date range.

        Parameters:
            from_date (str): The start date.
            to_date (str, optional): The end date.
            Format: ISO 8601 (YYYY-MM-DDTHH:MM:SS.sssZ)

        Returns:
             Invoice DTO: The transaction history.
        """
        Utils.validate_date(from_date)
        if to_date:
            Utils.validate_date(to_date)

        params = {"fromDate": from_date, "toDate": to_date, "limit": limit}
        response_data = self._get_request(Endpoints.invoices_history, params)
        invoices = [Invoice(**data) for data in response_data.get('invoices')]
        return InvoicesHistory(
            invoices=invoices,
            totalCount=response_data.get('totalCount')
        )

    def get_transactions(self, from_date: str, to_date: Optional[str] = None, limit: int = 100) -> TransactionsHistory:
        """
        Retrieves transaction history within a given date range.

        Parameters:
            from_date (str): The start date.
            to_date (str, optional): The end date.
            Format: ISO 8601 (YYYY-MM-DDTHH:MM:SS.sssZ)

        Returns:
             Transaction DTO: The transaction history.
        """
        Utils.validate_date(from_date)
        if to_date:
            Utils.validate_date(to_date)

        params = {"fromDate": from_date, "toDate": to_date, "limit": limit}
        response_data = self._get_request(Endpoints.transactions_history, params)
        transactions = [Transaction(**data) for data in response_data.get('transactions')]
        return TransactionsHistory(
            transactions=transactions,
            totalCount=response_data.get('totalCount')
        )


    def get_wallet(self, id: Optional[str] = None, order_id: Optional[str] = None) -> Wallet:
        """
        Retrieve wallet information based on wallet ID or order ID.

        Args:
            id (Optional[str]): The ID of the wallet to retrieve.
            order_id (Optional[str]): The order ID associated with the wallet to retrieve.

        Returns:
            Wallet: An instance of the Wallet class containing the wallet information.

        Raises:
            ValueError: If neither `id` nor `order_id` is provided.
        """
        if id:
            params = {"id": id}
        elif order_id:
            params = {"orderId": order_id}
        response_data = self._get_request(Endpoints.wallet, params)
        return Wallet(**response_data)


    def create_invoice(
        self,
        order_id: str,
        description: str,
        amount: str,
        currency: str,
        fee_included: bool,
        accuracy: str,
        discount: str,
    ) -> Invoice:
        """
        Creates a new invoice.

        Parameters:
            order_id (str): The order ID.
            description (str): The invoice description.
            amount (str): The amount.
            currency (str): The currency.
            fee_included (bool): Whether the fee is included in the amount.
            accuracy (str): The accuracy level.
            discount (str): The discount.

        Returns:
             Invoice DTO: The created invoice data.
        """

        data = {
            "orderId": order_id,
            "description": description,
            "amount": amount,
            "currency": currency,
            "feeIncluded": fee_included,
            "accuracy": accuracy,
            "discount": discount,
        }
        return self._post_request(Endpoints.invoice, json=data)

    def create_wallet(self, order_id: str) -> Wallet:
        """
        Creates a new wallet.

        Parameters:
            order_id (str): The order ID.

        Returns:
             Wallet DTO: The created wallet data.
        """

        data = {
            "orderId": order_id
        }
        return self._post_request(Endpoints.wallet, json=data)


from pydantic import BaseModel, Field
from typing import List, Optional


class Address(BaseModel):
    chain: str
    token: str
    method_name: str = Field(alias='methodName')
    address: str
    token_rate: Optional[str] = Field(alias='tokenRate', default=None)


class FiatDetails(BaseModel):
    name: str
    amount: str
    payed_amount: str = Field(alias="payedAmount")
    fee_rate: str = Field(alias="feeRate")
    bank: str
    requisites: str
    card_owner: str = Field(alias="cardOwner")


class Invoice(BaseModel):
    id: str
    order_id: str = Field(alias="orderId")
    payed_amount: str = Field(alias="payedAmount")
    fee_included: bool = Field(alias="feeIncluded")
    accuracy: str
    discount: str
    fee_rate: str = Field(alias="feeRate")
    created_at: str = Field(alias="createdAt")
    expired_at: str = Field(alias="expiredAt")
    status: str
    addresses: List[Address]
    description: str
    amount: Optional[str] = Field(default="0")
    fiat_details: Optional[List[FiatDetails]] = Field(alias="FiatDetails", default=None)

    class Config:
        populate_by_name = True



from pydantic import BaseModel

from chiefpay.types.invoice import Invoice
from chiefpay.types.transaction import Transaction


class NotificationTransaction(BaseModel):
    type: str = "transaction"
    transaction: Transaction


class NotificationInvoice(BaseModel):
    type: str = "invoice"
    invoice: Invoice



import socketio

from chiefpay.constants import BASE_URL
from chiefpay.exceptions import SocketError
from chiefpay.socket.base import BaseSocketClient


class AsyncSocketClient(BaseSocketClient):
    """
    Client for interacting with the payment system via WebSockets (asynchronous).
    """
    def __init__(self, api_key: str, base_url: str = BASE_URL):
        super().__init__(api_key, base_url)
        self.sio = socketio.AsyncClient()
        self._setup_event_handlers()

    def _setup_event_handlers(self):
        @self.sio.event
        async def connect():
            print("Connected to Socket.IO server")

        @self.sio.event
        async def disconnect():
            print("Disconnected from Socket.IO server")

        @self.sio.event
        async def rates(data):
            self.rates = data
            if self.on_rates:
                await self.on_rates(data)

        @self.sio.event
        async def notification(data):
            if self.on_notification:
                await self.on_notification(data)

    async def connect(self):
        """
        Asynchronously connects to the Socket.IO server.

        Raises:
            SocketError: If the connection fails.
        """
        try:
            await self.sio.connect(
                self.base_url,
                headers={"X-Api-Key": self.api_key},
                socketio_path=self.PATH,
            )
        except Exception as e:
            raise SocketError(f"Failed to connect to Socket.IO server: {e}")

    async def disconnect(self):
        """
        Asynchronously disconnects from the Socket.IO server.
        """
        await self.sio.disconnect()


    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()


__all__ = (
    'Invoice',
    'Address',
    'FiatDetails',
    'Wallet',
    'Rate',
    'Transaction',
    'NotificationTransaction',
    'NotificationInvoice',
    'TransactionsHistory',
    'InvoicesHistory'
)

from chiefpay.types.invoice import Invoice, Address, FiatDetails
from chiefpay.types.history import TransactionsHistory, InvoicesHistory
from chiefpay.types.wallet import Wallet
from chiefpay.types.rates import Rate
from chiefpay.types.transaction import Transaction
from chiefpay.types.notification import NotificationTransaction, NotificationInvoice


from typing import Callable, Union
from chiefpay.base import BaseClient
from chiefpay.constants import *
from chiefpay.types import Rate
from chiefpay.types.notification import NotificationInvoice, NotificationTransaction


class BaseSocketClient(BaseClient):
    """
    Base class for interacting with the payment system via WebSockets.
    """
    PATH = Endpoints.socket.value

    def __init__(self, api_key: str, base_url: str = BASE_URL):
        """
        Initializes the socket client.

        Parameters:
            api_key (str): API key for authentication.
            base_url (str): Base URL for the API endpoints.
        """
        super().__init__(api_key, base_url)
        self.rates: list[Rate] = None
        self.on_rates = None
        self.on_notification = None

    def _init_session(self):
        return None

    def set_on_notification(self, callback: Callable[[Union[NotificationInvoice, NotificationTransaction]], None]):
        """
        Sets a callback function to handle incoming notifications.

        Parameters:
            callback (function): A function that takes one argument (the notification data)
                                 and handles it appropriately.
        """
        self.on_notification = callback

    def set_on_rates(self, callback: Callable[[Union[NotificationInvoice, NotificationTransaction]], None]):
        """
        Sets a callback function to handle incoming rates updates.

        Parameters:
            callback (function): A function that takes one argument (the rates data)
                                and handles it appropriately.
        """
        self.on_rates = callback

    def get_latest_rates(self) -> list[Rate] | None:
        """
        Retrieves the latest exchange rates.

        Returns:
            dict: The latest exchange rates.
        """
        return self.rates


__all__ = (
    'Client',
    'ChiefPayClient',
    'AsyncChiefPayClient',
    'AsyncClient',
    'SocketClient',
    'AsyncSocketClient'
)


from chiefpay.client import Client
from chiefpay.async_client import AsyncClient
from chiefpay.socket import SocketClient, AsyncSocketClient
from chiefpay.classes import ChiefPayClient, AsyncChiefPayClient
